//	FileLoader.cpp - Loads and saves from/to X3D files.
//	----------------------------------------------------------------------------
//	This file is part of Motion Composition Tool, a simple application for
//	creating simple abstract animations.
//	Copyright (c) 2007 Niall Moody.
//
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program; if not, write to the Free Software
//	Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//	----------------------------------------------------------------------------

#include "FileLoader.h"
#include "Window.h"
#include "StringParser.h"
#include "DrawingPanel.h"

/*#ifdef WIN32
#undef Rectangle
#include "Windows.h"
#endif*/
#include <iostream>
#include <sstream>

using namespace std;

//------------------------------------------------------------------------------
FileLoader::FileLoader(Window *parent):
FileBasedDocument(T(".x3d"),
				  T("*.x3d"),
				  T("Choose a file to open"),
				  T("Choose a file to save")),
window(parent)
{

}

//------------------------------------------------------------------------------
FileLoader::~FileLoader()
{
	
}

//------------------------------------------------------------------------------
const String FileLoader::getDocumentTitle()
{
	return T("MotionCompositioTool Document");
}

//------------------------------------------------------------------------------
const String FileLoader::loadDocument(const File& file)
{
	String retval;
	XmlDocument doc(file);
	XmlElement *tempEl = doc.getDocumentElement();
	XmlElement *tempEl2;
	XmlElement *tempEl3;
	XmlElement *tempEl4;
	int coloursFound = 0;
	int keysFound = 0;
	int keyValuesFound = 0;
	bool lengthFound = false;

	if(!tempEl)
		retval = doc.getLastParseError();
	else if(tempEl->hasTagName(T("X3D")))
	{
		tempEl2 = tempEl->getChildByName(T("Scene"));

		if(tempEl2)
		{
			tempEl3 = tempEl2->getFirstChildElement();
			while(tempEl3 &&
				  ((coloursFound != 4) ||
				   (keysFound != 4) ||
				   (keyValuesFound != 4) ||
				   !lengthFound))
			{
				if(tempEl3)
				{
					//Colours
					if(tempEl3->hasTagName(T("Transform")) &&
					   (coloursFound < 4))
					{
						tempEl4 = tempEl3->getChildByName(T("IndexedFaceSet"));

						if(tempEl4)
						{
							tempEl4 = tempEl4->getChildByName(T("ColorRGBA"));

							if(tempEl4)
							{
								StringParser parser((const char *)(tempEl4->getStringAttribute(T("color"))));

								if(parser.getNumFourFloats() > 0)
								{
									FourFloats tempf = parser.getFourFloats();
									Colour tempCol((uint8)(tempf.x * 255.0f),
												   (uint8)(tempf.y * 255.0f),
												   (uint8)(tempf.z * 255.0f),
												   (uint8)(tempf.r * 255.0f));

									window->setPolygonColour(coloursFound,
															 tempCol);
									++coloursFound;
								}
							}
						}
					}
					//Polygon positions.
					else if(tempEl3->hasTagName(T("CoordinateInterpolator")) &&
							(keysFound < 4))
					{
						int i, j;
						list<Polygon> tempList;
						StringParser keyParser((const char *)(tempEl3->getStringAttribute(T("key"))));
						StringParser valueParser((const char *)(tempEl3->getStringAttribute(T("keyValue"))));

						if(keyParser.getNumFloats() == valueParser.getNumEightFloats())
						{
							for(i=0;i<keyParser.getNumFloats();++i)
							{
								Polygon tempPoly;
								EightFloats tempEight = valueParser.getEightFloats(i);

								tempPoly.time = keyParser.getFloat(i);
								for(j=0;j<4;++j)
								{
									tempPoly.x[j] = tempEight.x[j];
									tempPoly.y[j] = tempEight.y[j];
								}

								tempList.push_back(tempPoly);
							}

							window->setPolygonList(keysFound, tempList);
							++keysFound;
						}
						else
						{
							retval = T("number of keys does not match number of keyValues");
							break;
						}
					}
					//Length.
					else if(tempEl3->hasTagName(T("TimeSensor")) && !lengthFound)
					{
						StringParser parser((const char *)tempEl3->getStringAttribute(T("cycleInterval")));

						if(parser.getNumFloats() > 0)
						{
							window->setLength(parser.getFloat());
							lengthFound = true;
						}
					}
				}
				tempEl3 = tempEl3->getNextElement();
			}
		}
		else
			retval = T("No Scene node in file.");

		delete tempEl;
	}
	else
	{
		retval = T("Not an X3D file.");

		delete tempEl;
	}

	if(coloursFound < 4)
		retval = T("Not enough colours found.");
	else if(keysFound < 4)
		retval = T("Not enough polygons found.");
	else if(!lengthFound)
		retval = T("No length parameter found.");

	return retval;
}

//------------------------------------------------------------------------------
const String FileLoader::saveDocument(const File& file)
{
	String retval;
	FileOutputStream *outStream;

	file.deleteFile();

	outStream = file.createOutputStream();

	if(outStream)
	{
		*outStream << outFile1;

		//Write colour 1.
		{
			stringstream tempConverter;
			Colour tempCol = window->getPolygonColour(0);

			tempConverter << tempCol.getFloatAlpha() << " ";
			tempConverter << tempCol.getFloatRed() << " ";
			tempConverter << tempCol.getFloatGreen() << " ";
			tempConverter << tempCol.getFloatBlue();

			*outStream << tempConverter.str().c_str();
		}

		*outStream << outFile2;

		//Write key and keyValue 1.
		{
			const list<Polygon>& polys = window->getPolygonList(0);

			//Write key.
			{
				stringstream tempConverter;
				list<Polygon>::const_iterator it;

				for(it=polys.begin();it!=polys.end();++it)
					tempConverter << it->time << " ";

				*outStream << tempConverter.str().c_str();
			}

			*outStream << outFile3;

			//Write keyValue.
			{
				int i;
				stringstream tempConverter;
				list<Polygon>::const_iterator it;

				for(it=polys.begin();it!=polys.end();++it)
				{
					for(i=0;i<4;++i)
					{
						tempConverter << it->x[i] << " ";
						tempConverter << it->y[i] << " ";
						tempConverter << "0 ";
					}
				}

				*outStream << tempConverter.str().c_str();
			}
		}

		*outStream << outFile4;

		//Write colour 2.
		{
			stringstream tempConverter;
			Colour tempCol = window->getPolygonColour(1);

			tempConverter << tempCol.getFloatAlpha() << " ";
			tempConverter << tempCol.getFloatRed() << " ";
			tempConverter << tempCol.getFloatGreen() << " ";
			tempConverter << tempCol.getFloatBlue();

			*outStream << tempConverter.str().c_str();
		}

		*outStream << outFile5;

		//Write key and keyValue 2.
		{
			const list<Polygon>& polys = window->getPolygonList(1);

			//Write key.
			{
				stringstream tempConverter;
				list<Polygon>::const_iterator it;

				for(it=polys.begin();it!=polys.end();++it)
					tempConverter << it->time << " ";

				*outStream << tempConverter.str().c_str();
			}

			*outStream << outFile6;

			//Write keyValue.
			{
				int i;
				stringstream tempConverter;
				list<Polygon>::const_iterator it;

				for(it=polys.begin();it!=polys.end();++it)
				{
					for(i=0;i<4;++i)
					{
						tempConverter << it->x[i] << " ";
						tempConverter << it->y[i] << " ";
						tempConverter << "0 ";
					}
				}

				*outStream << tempConverter.str().c_str();
			}
		}

		*outStream << outFile7;

		//Write colour 3.
		{
			stringstream tempConverter;
			Colour tempCol = window->getPolygonColour(2);

			tempConverter << tempCol.getFloatAlpha() << " ";
			tempConverter << tempCol.getFloatRed() << " ";
			tempConverter << tempCol.getFloatGreen() << " ";
			tempConverter << tempCol.getFloatBlue();

			*outStream << tempConverter.str().c_str();
		}

		*outStream << outFile8;

		//Write key and keyValue 3.
		{
			const list<Polygon>& polys = window->getPolygonList(2);

			//Write key.
			{
				stringstream tempConverter;
				list<Polygon>::const_iterator it;

				for(it=polys.begin();it!=polys.end();++it)
					tempConverter << it->time << " ";

				*outStream << tempConverter.str().c_str();
			}

			*outStream << outFile9;

			//Write keyValue.
			{
				int i;
				stringstream tempConverter;
				list<Polygon>::const_iterator it;

				for(it=polys.begin();it!=polys.end();++it)
				{
					for(i=0;i<4;++i)
					{
						tempConverter << it->x[i] << " ";
						tempConverter << it->y[i] << " ";
						tempConverter << "0 ";
					}
				}

				*outStream << tempConverter.str().c_str();
			}
		}

		*outStream << outFile10;

		//Write colour 4.
		{
			stringstream tempConverter;
			Colour tempCol = window->getPolygonColour(3);

			tempConverter << tempCol.getFloatAlpha() << " ";
			tempConverter << tempCol.getFloatRed() << " ";
			tempConverter << tempCol.getFloatGreen() << " ";
			tempConverter << tempCol.getFloatBlue();

			*outStream << tempConverter.str().c_str();
		}

		*outStream << outFile11;

		//Write key and keyValue 4.
		{
			const list<Polygon>& polys = window->getPolygonList(3);

			//Write key.
			{
				stringstream tempConverter;
				list<Polygon>::const_iterator it;

				for(it=polys.begin();it!=polys.end();++it)
					tempConverter << it->time << " ";

				*outStream << tempConverter.str().c_str();
			}

			*outStream << outFile12;

			//Write keyValue.
			{
				int i;
				stringstream tempConverter;
				list<Polygon>::const_iterator it;

				for(it=polys.begin();it!=polys.end();++it)
				{
					for(i=0;i<4;++i)
					{
						tempConverter << it->x[i] << " ";
						tempConverter << it->y[i] << " ";
						tempConverter << "0 ";
					}
				}

				*outStream << tempConverter.str().c_str();
			}
		}

		*outStream << outFile13;

		//Write length.
		{
			*outStream << window->getLength();
		}

		*outStream << outFile14;

		outStream->flush();
		delete outStream;
	}
	else
		retval = T("Could not create output stream.");

	return retval;
}

//------------------------------------------------------------------------------
const File FileLoader::getLastDocumentOpened()
{
	return lastFile;
}

//------------------------------------------------------------------------------
void FileLoader::setLastDocumentOpened(const File& file)
{
	lastFile = file;
}

//------------------------------------------------------------------------------
const char * const FileLoader::outFile1 = "<?xml version=\"1.0\" encoding=\"utf-8\"?>\n"\
	"<!DOCTYPE X3D>\n"\
	"<X3D profile=\"Full\">\n"\
	"	<Scene>\n"\
	"		<NavigationInfo type=\"ANY\"/>\n"\
	" \n"\
	"		<Transform translation=\"-2 -1.5 4.501\" scale=\"4 3 1\">\n"\
	"			<IndexedFaceSet coordIndex=\"0 1 2 3 -1\" colorIndex=\"0 0 0 0 -1\">\n"\
	"				<Coordinate DEF=\"poly1\" point=\"0 0 0 0 1 0 1 0 0 1 -1 0\"/>\n"\
	"				<ColorRGBA color=\"";

const char * const FileLoader::outFile2 = "\"/>\n"\
	"			</IndexedFaceSet>\n"\
	"		</Transform>\n"\
	"		<CoordinateInterpolator DEF=\"interp1\" key=\"";

const char * const FileLoader::outFile3 = "\" keyValue=\"";

const char * const FileLoader::outFile4 = "\"/>\n"\
	" \n"\
	"		<Transform translation=\"-2 -1.5 4.502\" scale=\"4 3 1\">\n"\
	"			<IndexedFaceSet coordIndex=\"0 1 2 3 -1\" colorIndex=\"0 0 0 0 -1\">\n"\
	"				<Coordinate DEF=\"poly2\" point=\"0 0 0 0 1 0 1 0 0 1 -1 0\"/>\n"\
	"				<ColorRGBA color=\"";

const char * const FileLoader::outFile5 = "\"/>\n"\
	"			</IndexedFaceSet>\n"\
	"		</Transform>\n"\
	"		<CoordinateInterpolator DEF=\"interp2\" key=\"";

const char * const FileLoader::outFile6 = "\" keyValue=\"";

const char * const FileLoader::outFile7 = "\"/>\n"\
	" \n"\
	"		<Transform translation=\"-2 -1.5 4.503\" scale=\"4 3 1\">\n"\
	"			<IndexedFaceSet coordIndex=\"0 1 2 3 -1\" colorIndex=\"0 0 0 0 -1\">\n"\
	"				<Coordinate DEF=\"poly3\" point=\"0 0 0 0 1 0 1 0 0 1 -1 0\"/>\n"\
	"				<ColorRGBA color=\"";

const char * const FileLoader::outFile8 = "\"/>\n"\
	"			</IndexedFaceSet>\n"\
	"		</Transform>\n"\
	"		<CoordinateInterpolator DEF=\"interp3\" key=\"";

const char * const FileLoader::outFile9 = "\" keyValue=\"";

const char * const FileLoader::outFile10 = "\"/>\n"\
	" \n"\
	"		<Transform translation=\"-2 -1.5 4.504\" scale=\"4 3 1\">\n"\
	"			<IndexedFaceSet coordIndex=\"0 1 2 3 -1\" colorIndex=\"0 0 0 0 -1\">\n"\
	"				<Coordinate DEF=\"poly4\" point=\"0 0 0 0 1 0 1 0 0 1 -1 0\"/>\n"\
	"				<ColorRGBA color=\"";

const char * const FileLoader::outFile11 = "\"/>\n"\
	"			</IndexedFaceSet>\n"\
	"		</Transform>\n"\
	"		<CoordinateInterpolator DEF=\"interp4\" key=\"";

const char * const FileLoader::outFile12 = "\" keyValue=\"";

const char * const FileLoader::outFile13 = "\"/>\n"\
	" \n"\
	"		<TimeSensor DEF=\"timer\" cycleInterval=\"";

const char * const FileLoader::outFile14 = "\" loop=\"false\"/>\n"\
	" \n"\
	"		<ROUTE fromNode=\"timer\" fromField=\"fraction_changed\" toNode=\"interp1\" toField=\"set_fraction\"/>\n"\
	"		<ROUTE fromNode=\"timer\" fromField=\"fraction_changed\" toNode=\"interp2\" toField=\"set_fraction\"/>\n"\
	"		<ROUTE fromNode=\"timer\" fromField=\"fraction_changed\" toNode=\"interp3\" toField=\"set_fraction\"/>\n"\
	"		<ROUTE fromNode=\"timer\" fromField=\"fraction_changed\" toNode=\"interp4\" toField=\"set_fraction\"/>\n"\
	"		<ROUTE fromNode=\"interp1\" fromField=\"value_changed\" toNode=\"poly1\" toField=\"point\"/>\n"\
	"		<ROUTE fromNode=\"interp2\" fromField=\"value_changed\" toNode=\"poly2\" toField=\"point\"/>\n"\
	"		<ROUTE fromNode=\"interp3\" fromField=\"value_changed\" toNode=\"poly3\" toField=\"point\"/>\n"\
	"		<ROUTE fromNode=\"interp4\" fromField=\"value_changed\" toNode=\"poly4\" toField=\"point\"/>\n"\
	"	</Scene>\n"\
	"</X3D>\n";
